// devFanuc.cc
extern "C" {
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <semLib.h>
#include <tickLib.h>
#include <taskLib.h>
 
#include "dbAccess.h"
#include "dbDefs.h"
#include "link.h"
#include "epicsPrint.h"
#include "dbCommon.h"
#include "biRecord.h"
#include "waveformRecord.h"
#include "recSup.h"
#include "dbEvent.h"
}

#include "Message.h"
#include "Char8ArrayMessage.h"
#include "Int32Message.h"
#include "DevMpf.h"

class FanucWf : public DevMpf
{
public:
    FanucWf(dbCommon*,DBLINK*);
    long startIO(dbCommon* pr);
    long completeIO(dbCommon* pr,Message* m);
    static long dev_init(void*);
    int sendMonitorRequest();
    IOSCANPVT *scan_list;
    bool firstTime;
private:
    // disallow
    FanucWf();
    FanucWf(const FanucWf &);
    FanucWf& operator=(const FanucWf&);
};

MAKE_DSET(devWfFanuc,FanucWf::dev_init)

long FanucWf::dev_init(void* v)
{
    waveformRecord* pr = (waveformRecord*)v;
    FanucWf* pFanucWf = new FanucWf((dbCommon*)pr,&(pr->inp));
    return(pFanucWf->getStatus());
}

int FanucWf::sendMonitorRequest()
{
    Int32Message *message = new Int32Message;
    message->cmd = 1;
    int status = send(message);
    return(status);
}

FanucWf::FanucWf(dbCommon* pr,link* l)
: DevMpf(pr,l,true), scan_list(0), firstTime(true)
{
    waveformRecord* wf = (waveformRecord*)pr;
    wf->pact=TRUE; // prevent scanning
    if((wf->ftvl!=DBF_SHORT) && (wf->ftvl!=DBF_USHORT)) {
	epicsPrintf("%s devFanuc (init_record) Illegal data type\n",pr->name);
        return;
    }
    scan_list = (IOSCANPVT *)calloc(wf->nelm,sizeof(IOSCANPVT));
    for(unsigned int i=0; i<wf->nelm; i++) scanIoInit(&(scan_list[i]));
    setProcessOnConnect(true);
    sendMonitorRequest();
}

long FanucWf::startIO(dbCommon* )
{
    // This is called whenever connecton state changes
    // just try to issue  monitor request
    sendMonitorRequest();
    return(0);
}

long FanucWf::completeIO(dbCommon* pr,Message* m)
{
    waveformRecord* pwf = (waveformRecord*)pr;
    unsigned short  monitor_mask;
    monitor_mask = recGblResetAlarms(pwf);
    if(m->getType() != messageTypeChar8Array) {
	if(m->getType() != messageTypeInt32)
            epicsPrintf("FanucWf::completeIO illegal message. %s\n",
                pwf->name);
        recGblSetSevr(pr,READ_ALARM,INVALID_ALARM);
        return(0);
    }
    Char8ArrayMessage *message = (Char8ArrayMessage *)m;
    int32 rtnSize = message->getSize();
    // each two unsigned chars will be put into uint16
    unsigned int size = rtnSize/2;
    if(rtnSize%2 != 0 || size>pwf->nelm) {
	epicsPrintf("FanucWf::completeIO illegal message length. %s\n",
	    pwf->name);
	recGblSetSevr(pr,READ_ALARM,INVALID_ALARM);
	return(0);
    }
    unsigned char *rcvBuf = (unsigned char *)message->value;
    if(message->status==0)
    {
	uint16 *buf = (uint16 *)pwf->bptr;
	for(unsigned int i=0; i<size; i++) {
	    uint16 val = rcvBuf[i*2]<<8 | rcvBuf[i*2 + 1];
	    if((buf[i] != val) || firstTime) {
		monitor_mask |= (DBE_LOG|DBE_VALUE);
		scanIoRequest(scan_list[i]);
	        buf[i] = val;
	    }
	}
	firstTime = false;
	pwf->nord = size;
	pwf->udf = 0;
    }
    else {
        recGblSetSevr(pr,READ_ALARM,INVALID_ALARM);
	for(unsigned int i=0; i<pwf->nelm; i++) scanIoRequest(scan_list[i]);
    }
    delete message;
    if(monitor_mask) db_post_events(pwf,pwf->bptr,monitor_mask);
    return(0);
}

/*************** BI SUPPORT ***************/
class DevPvt
{
public:
    int word; // word offset in bit arra
    unsigned short bit;     // bit within word in bit array (pattern)
    waveformRecord *pwf;
};

static long bi_init_record(biRecord*);
static long bi_ioint_info(int cmd,dbCommon* pr,IOSCANPVT* iopvt);
static long bi_read(biRecord*);

extern "C" {
MPF_DSET devBiFanuc =
{5,NULL,NULL,bi_init_record,bi_ioint_info,bi_read,NULL};
};


static long bi_init_record(biRecord* pr)
{
    DevPvt* pvt = new DevPvt;
    pvt->word = -1;
    link *plink = &pr->inp;
    if(plink->type!=INST_IO)
    {
	epicsPrintf("%s devFanuc Illegal INP field\n",pr->name);
        pr->pact = TRUE;
        return(0);
    }
    instio *inst = (instio*)&(plink->value);
    char mystring[100];
    ::strcpy(mystring,inst->string);
    char* save;
    char *pv=strtok_r(mystring,"[",&save);
    char *bitpos=strtok_r(NULL,"]",&save);
    if(pv==0 || bitpos==0) {
	epicsPrintf("%s devFanuc Illegal INP field\n",pr->name);
        pr->pact = TRUE;
        return(0);
    }
    DBADDR db;
    if(dbNameToAddr(pv,&db)) {
	epicsPrintf("%s devFanuc Can't locate PV\n",pr->name);
        pr->pact = TRUE;
        return(0);
    }
    waveformRecord *pwf=(waveformRecord*)db.precord;
    int bit;
    if(sscanf(bitpos,"%d",&bit)<1) {
	epicsPrintf("%s devFanuc Illegal INP Bit field\n",pr->name);
	pr->pact = TRUE;
	return(0);
    }
    pvt->word=(bit-1)/(8*sizeof(short));
    pvt->bit=1<<((bit-1)%(8*sizeof(short)));
    if(pvt->word>(int)pwf->nelm) {
	epicsPrintf("%s devFanuc INP bit field too large\n",pr->name);
	pr->pact = TRUE;
	return(0);
    }
    pvt->pwf = pwf;
    pr->dpvt = (void *)pvt;
    return(0);
}

static long bi_ioint_info(int ,dbCommon* pr,IOSCANPVT* iopvt)
{
    DevPvt* pvt = (DevPvt*)pr->dpvt;
    if(!pvt) return(0);
    FanucWf *pFanucWf = (FanucWf *)pvt->pwf->dpvt;
    IOSCANPVT* scan_list = pFanucWf->scan_list;
    *iopvt = scan_list[pvt->word];
    return 0;
}

static long bi_read(biRecord* pr)
{
    DevPvt* pvt = (DevPvt*)pr->dpvt;
    if(!pvt) return(0);
    uint16 *data = (uint16 *)pvt->pwf->bptr;
    pr->rval = ((data[pvt->word])&(pvt->bit)) ? 1 : 0;
    recGblSetSevr(pr,pvt->pwf->stat,pvt->pwf->sevr);
    pr->udf=0;
    return 0; 
}
