// fanucServer.cc

// Author: Jim Kowalkowski
// Date: 6/95
// New Author: Marty Kraimer 
// Date: 8/98

extern "C" {
#include <vxWorks.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <taskLib.h>
#include <semLib.h>
}
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "SerialPort.h"


/* Notes:
   A client can send two Int32Messages to the fanucServer 
      cmd = 1 means start  monitoring
      cmd = 2 means stop monitoring
   Only one client at a time can monitor the fanuc
   A Int32 message is sent to the client if any failure occurs.
   A Char8ArrayMessage is sent to the client whenever a value change occurs
*/

static const int asciiDataSize = 200;
static const int readRetrys = 3;

enum {STX='s',ETX='t'};
enum readState {readIdle, readReading, readDone};

class Fanuc
{
public:
    Fanuc(const char *name,const char *portName, int timeout);
    bool getStartOk() const {return(startOk);}
    static void fanucServer(Fanuc *);
    static byteHandlerRC byteHandler(void*,unsigned char);
    void convert();
private:
    unsigned char convAsciiToBinary(char x);
    const char *serverName;
    MessageServer *pMessageServer;
    SerialPort* pSerialPort;
    bool startOk;
    Int32Message *preceive;
    int timeout;
    serialPortStatus status;
    readState state;
    int nextRead;
    unsigned char asciiData[asciiDataSize];
    unsigned char binaryData[asciiDataSize/2];
    int binaryDataLength;
    bool valueChange;
};

static char taskname[] = "fanuc";
int initFanucServer(char *serverName,char *portName,int timeout)
{
    Fanuc *pFanuc = new Fanuc(serverName,portName,timeout);
    if(!pFanuc->getStartOk()) return(0);
    int taskId = taskSpawn(taskname,100,VX_FP_TASK,2000,
	(FUNCPTR)Fanuc::fanucServer,(int)pFanuc,0,0,0,0,0,0,0,0,0);
    if(taskId==ERROR) 
	printf("%s fanucServer taskSpawn Failure\n",serverName);
    return(0);
}
	

Fanuc:: Fanuc(const char *name,const char *portName,int tout)
: serverName(name), pMessageServer(0), pSerialPort(0),
  startOk(false), preceive(0), timeout(tout),
  state(readDone), nextRead(0), binaryDataLength(0),
  valueChange(true)
{
    pSerialPort = SerialPort::bind(portName,this,byteHandler);
    if(pSerialPort==0) {
	printf("%s could not bind to SerialPort\n",name);
        return;
    }
    pMessageServer = new MessageServer(name);
    startOk = true;
}

void Fanuc::fanucServer(Fanuc *pFanuc)
{
    while(true) {
        pFanuc->pMessageServer->waitForMessage();
	while(true) {
	    Message *message = pFanuc->pMessageServer->receive();
	    if(message) {
                if(message->getType()!=messageTypeInt32) {
                    printf("%s illegal message type %d\n",
		        pFanuc->serverName, message->getType());
		    delete message;
		    continue;
		}
		if(pFanuc->preceive) delete pFanuc->preceive;
		Int32Message *pInt32= (Int32Message *)message;
		if(pInt32->cmd==2) { //stop monitoring
		    pFanuc->preceive = 0;
		} else if(pInt32->cmd!=1) {
		    pFanuc->preceive = 0;
		    printf("%s received illegal cmd %d\n",
			pFanuc->serverName,pInt32->cmd);
		} else {
		    if(pFanuc->preceive)
			printf("%s received additional client."
			"Replaces original\n",pFanuc->serverName);
		    pFanuc->preceive = pInt32;
		}
		// make sure next client gets first message
		pFanuc->valueChange = true;
            }
	    if(pFanuc->preceive==0) break;
	    int sendStatus = 0;
	    pFanuc->status = serialPortFailure;
	    for(int nextTry=0; nextTry<readRetrys; nextTry++) {
		pFanuc->state = readIdle;
		pFanuc->status = pFanuc->pSerialPort->read(pFanuc->timeout);
		if(pFanuc->status==serialPortSuccess) break;
	    }
	    if(pFanuc->status==serialPortSuccess) pFanuc->convert();
	    if(pFanuc->status!=serialPortSuccess) {
		printf("%s status %d\n",pFanuc->serverName,pFanuc->status);
		Int32Message *psend = (Int32Message *)pFanuc->pMessageServer->
			allocReplyMessage(pFanuc->preceive,messageTypeInt32);
		psend->status = -1;
                sendStatus = pFanuc->pMessageServer->reply(psend);
	    } else if(pFanuc->valueChange){
		pFanuc->valueChange = false;
                Char8ArrayMessage *psend =
		    (Char8ArrayMessage *)pFanuc->pMessageServer->
                    allocReplyMessage(pFanuc->preceive,messageTypeChar8Array);
		psend->allocValue(pFanuc->binaryDataLength);
		::memcpy(psend->value,pFanuc->binaryData,
		    pFanuc->binaryDataLength);
		psend->setSize(pFanuc->binaryDataLength);
		psend->status = 0;
                sendStatus = pFanuc->pMessageServer->reply(psend);
	    }
	    if(sendStatus!=0) {
		delete pFanuc->preceive;
		pFanuc->preceive = 0;
	    }
	}
    }
    return;
}

/*
    packet formats:
    <STX><data><data>.........<BCC><BCC><ETX>

    Each <data> and <BCC> is ascii 0,...,9 or A,...F or a,...f
    each 2 <data> or <BCC> represents an 8 bit binary value

    BCC calculation:
    ~( lower 8 bits of sum of all application user data (user stuff) ) + 1
    MARTY: I think this comment is wrong

    This task reads data from the GE Fanuc.  The Fanuc justs continually
    pukes (Jim's word) out it's digital I/O data in 8 bit bytes
    as two ascii characters in hex notation.

*/
byteHandlerRC Fanuc::byteHandler(void* v, unsigned char data)
{
    Fanuc* pFanuc = (Fanuc*)v;
    switch(pFanuc->state) {
    case readIdle:
        if(data==STX) {
	    pFanuc->state=readReading;
	    pFanuc->nextRead = 0; 
	}
        break;
    case readReading:
	if(data==ETX) {
	    pFanuc->state = readDone;
	    return(byteHandlerEndRead);
	} else {
	    if(pFanuc->nextRead>=asciiDataSize) {
		return(byteHandlerError);
	    } else {
		pFanuc->asciiData[pFanuc->nextRead++]=data;
	    }
	}
	break;
    case readDone:
	// just ignore all bytes until state = stateIdle
	break;
    default:
        break;
    }
    return(byteHandlerOK);
}

void Fanuc::convert()
{
    // packet should be even number of bytes
    if(nextRead%2 != 0) {
	printf("%s Fanuc::convert Illegal number of characters\n",serverName);
	status = serialPortFailure;
	return;
    }
    binaryDataLength = 0;
    long sum = 0;
    int checksumInd = nextRead - 2;
    for(int nextIn=0; nextIn < checksumInd; nextIn+=2) {
	unsigned char val = convAsciiToBinary(asciiData[nextIn])<<4
	    | convAsciiToBinary(asciiData[nextIn + 1]);
	sum += (long)val;
	if(binaryData[binaryDataLength]!=val) valueChange = true;
	binaryData[binaryDataLength++] = val;
    }
    // get checksum
    unsigned char val = convAsciiToBinary(asciiData[checksumInd]) <<4
	| convAsciiToBinary(asciiData[checksumInd+1]);
    long csum = sum&0x000000ff;
    if(val!=csum) {
	printf("%s Fanuc::convert checksum error\n",serverName);
	state = readIdle;
	status = serialPortFailure;
	return;
    }
    state = readIdle;
}

unsigned char Fanuc::convAsciiToBinary(char x)
{
    unsigned char val;

    if(x>='0' && x<='9') val=(x-'0');
    else if(x>='A' && x<='F') val=(x-'A'+10);
    else if(x>='a' && x<='f') val=(x-'a'+10);
    else return 0;
    return val;
}
